/*  image.c
 *
 *
 *  Image processing functions of wxapt application
 */

/*
 *  wxapt: An application to decode APT signals from weather
 *  satellites and produce an image of the weather.
 *
 *  Copyright (C) 2001  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include "wxapt.h"

/* Weather satellite type (NOAA/Meteor/etc) */
extern int gbl_sat_type;

/* Satellite type names */
extern char *gbl_sat_names[];

/* Required position of sync train in line buffer */
extern int gbl_sync_ref;

/* List of required sync position references */
extern int gbl_sync_refs[];

/* Pointer to samples file */
extern FILE *gbl_sample_fp;

/* Audio/DSP samples buffer */
unsigned char gbl_line_buffer[BUFFER_SIZE];

/* Index to audio samples buffer */
extern int gbl_buffer_idx;

extern int
  gbl_action, /* Action to carry out according to options   */
  gbl_stop;   /* Signals end of operation on user interrupt */

/*------------------------------------------------------------------------*/

/*  Detect_NOAA_Sync()
 *
 *  Detects the Channel A sync of a NOAA APT signal.
 *  This is 7 cycles at 1040 Hz on the 2.4 KHz sub-carrier.
 */

  int
Detect_NOAA_Sync( int search_limit )
{
  /* FIFO Buffer for ch A sync detection */
  static unsigned char sync_fifo[NOAA_FIFO_SIZE];

  int
	sample_a,      /* First of the pair of carrier samples  */
	sample_b,      /* Second of the pair of carrier samples */
	sync_cnt,      /* Counter used in detecting sync train position  */
	carrier_val,   /* Instantaneous value of 2.4 Khz sub-carrier     */
	sync_val,      /* Value produced by sync detection algorithm     */
	sync_max,      /* Max value produced by sync detection algorithm */
	sync_ave,      /* Average value of carrier samples in sync fifo  */
	sync_pos,      /* Position in sample buffer where sync o/p peaks */
	peak_idx,      /* Index used in searching for peak in sync value */
	fifo_idx,      /* Index for sync detection fifo */
	idx;           /* Index for various loops etc.  */


  /* Move through buffer up to search_limit, looking   */
  /* for sync detector peak. On exit, the buffer index */
  /* should be at the end of A sync pulse sequence.    */

  sync_max = fifo_idx = 0;
  sync_pos = gbl_buffer_idx;

  for( peak_idx = 0; peak_idx < search_limit-1; peak_idx++ )
  {
	/* Read first of the pair of carrier samples */
	sample_a = gbl_line_buffer[gbl_buffer_idx];

	/* If end of samples block reached, return error */
	if( ++gbl_buffer_idx >= BLOCK_SIZE )
	  return(-1);

	/* Read second of the pair of carrier samples */
	sample_b = gbl_line_buffer[gbl_buffer_idx];

	/* This is the instantaneous magnitude of the 2.4 Khz  */
	/* subcarrier, calculated as the scalar magnitude of   */
	/* two consecutive audio samples at 90 deg phase angle */
	carrier_val = Scalar_Mag( sample_a, sample_b );

	/* Store 63 samples in fifo in cyclic manner */
	if( ++fifo_idx == NOAA_FIFO_SIZE )
	  fifo_idx = 0;
	sync_fifo[fifo_idx] = carrier_val;

	/* Evaluate the average of the samples in fifo, to */
	/* be used as a zero reference in summation below. */
	sync_val = 0;
	for( idx = 0; idx < NOAA_FIFO_SIZE; idx++ )
	  sync_val += sync_fifo[idx];
	sync_ave = sync_val/NOAA_FIFO_SIZE;

	/* Add and subtract alternate 4 sample groups in the   */
	/* sync accumulator, simulating synchronous detection. */
	/* Drop one sample between each pair of groups as      */
	/* there are ~9 carrier sample values per sync cycle.  */
	/* Their average value is used as a zero reference.    */
	sync_val = sync_cnt = 0;
	while( sync_cnt < NOAA_FIFO_SIZE )
	{
	  /* Add 4 carrier samples into accumulator */
	  for( idx = 0; idx < 4; idx++ )
	  {
		sync_val += sync_fifo[fifo_idx] - sync_ave;
		if( ++fifo_idx == NOAA_FIFO_SIZE )
		  fifo_idx = 0;
		sync_cnt++;
	  }

	  /* Drop one sample between two 4-sample groups  */
	  /* This seems the easiest way to accomodate ~9  */
	  /* carrier samples per sync cycle of sync train */
	  if( ++fifo_idx == NOAA_FIFO_SIZE )
		fifo_idx = 0;
	  sync_cnt++;

	  for( idx = 0; idx < 4; idx++ )
	  {
		/* Subtract 4 carrier samples from accumulator */
		sync_val -= sync_fifo[fifo_idx] - sync_ave;
		if( ++fifo_idx == NOAA_FIFO_SIZE )
		  fifo_idx = 0;
		sync_cnt++;
	  }

	} /* End of while( sync_cnt < NOAA_FIFO_SIZE ) */

	/* Record position in buffer where sync detector peaks */
	if( (sync_val > sync_max) &&
		(sync_val > NOAA_SYNC_LOWER_THRESH) &&
		(sync_val < NOAA_SYNC_UPPER_THRESH) )
	{
	  sync_max = sync_val;
	  sync_pos = gbl_buffer_idx;
	}

  } /* End of for( peak_idx = 0; peak_idx < search_limit-1; peak_idx++ ) */

  /* Setup buffer index to max sync position */
  gbl_buffer_idx = sync_pos;

  return(0);

} /* End of Detect_NOAA_Sync() */

/*------------------------------------------------------------------------*/

/*  Detect_Meteor_Sync()
 *
 *  Detects the sync train of a Meteor type APT signal.
 *  This is 5 cycles at 1200 Hz on the 2.4 KHz sub-carrier.
 */

  int
Detect_Meteor_Sync( int search_limit )
{
  /* FIFO Buffer for sync detection */
  static unsigned char sync_fifo[METEOR_FIFO_SIZE];

  int
	sample_a,      /* First of the pair of carrier samples  */
	sample_b,      /* Second of the pair of carrier samples */
	sync_cnt,      /* Counter used in detecting sync train position  */
	carrier_val,   /* Instantaneous value of 2.4 Khz sub-carrier     */
	sync_val,      /* Value produced by sync detection algorithm     */
	sync_max,      /* Max value produced by sync detection algorithm */
	sync_ave,      /* Average value of carrier samples in sync fifo  */
	sync_pos,      /* Position in sample buffer where sync o/p peaks */
	peak_idx,      /* Index used in searching for peak in sync value */
	fifo_idx,      /* Index for sync detection fifo */
	idx;           /* Index for various loops etc.  */


  /* Move through buffer until sync detector  */
  /* peaks. On exit, the buffer index should  */
  /* be at the end of the sync pulse sequence */
  sync_max = fifo_idx = 0;
  sync_pos = gbl_buffer_idx;

  for( peak_idx = 0; peak_idx < search_limit-1 ; peak_idx++ )
  {
	/* Read first of the pair of carrier samples */
	sample_a = gbl_line_buffer[gbl_buffer_idx];

	/* If end of samples blocke reached, return error */
	if( ++gbl_buffer_idx >= BLOCK_SIZE )
	  return(-1);

	/* Read second of the pair of carrier samples */
	sample_b = gbl_line_buffer[gbl_buffer_idx];

	/* This is the instantaneous magnitude of the 2.4 Khz  */
	/* subcarrier, calculated as the scalar magnitude of   */
	/* two consecutive audio samples at 90 deg phase angle */
	carrier_val = Scalar_Mag( sample_a, sample_b );

	/* Store 40 samples in fifo in cyclic manner */
	if( ++fifo_idx == METEOR_FIFO_SIZE )
	  fifo_idx = 0;
	sync_fifo[fifo_idx] = carrier_val;

	/* Evaluate the average of the samples in fifo, to */
	/* be used as a zero reference in summation below. */
	sync_val = 0;
	for( idx = 0; idx < METEOR_FIFO_SIZE; idx++ )
	  sync_val += sync_fifo[idx];
	sync_ave = sync_val/METEOR_FIFO_SIZE;

	/* Subtract 5x3-sample groups and add 5x5-sample groups */
	/* to the sync accumulator alternately, simulating      */
	/* synchromous detection. The sync train appears to be  */
	/* 5 cycles at 1200 Hz with a 2/1 mark/space ratio.     */
	/* Their average value is used as a zero reference.     */
	sync_val = sync_cnt = 0;
	while( sync_cnt < METEOR_FIFO_SIZE )
	{
	  /* Subtract 3 carrier samples from accumulator */
	  for( idx = 0; idx < 3; idx++ )
	  {
		sync_val -= sync_fifo[fifo_idx] - sync_ave;
		if( ++fifo_idx == METEOR_FIFO_SIZE )
		  fifo_idx = 0;
		sync_cnt++;
	  }

	  for( idx = 0; idx < 5; idx++ )
	  {
		/* Add 5 carrier samples into accumulator */
		sync_val += sync_fifo[fifo_idx] - sync_ave;
		if( ++fifo_idx == METEOR_FIFO_SIZE )
		  fifo_idx = 0;
		sync_cnt++;
	  }

	} /* End of while( sync_cnt < METEOR_FIFO_SIZE ) */

	/* Record position in buffer where sync detector peaks */
	if( (sync_val > sync_max) &&
		(sync_val > METEOR_SYNC_LOWER_THRESH) &&
		(sync_val < METEOR_SYNC_UPPER_THRESH) )
	{
	  sync_max = sync_val;
	  sync_pos = gbl_buffer_idx;
	}

  } /* End of for( peak_idx = 0; peak_idx < search_limit-1; peak_idx++ ) */

  /* Setup buffer index to max sync position */
  gbl_buffer_idx = sync_pos;

  return(0);

} /* End of Detect_Meteor_Sync() */

/*------------------------------------------------------------------------*/

/*  Detect_Sync_Train()
 *
 *  Detects sync train position in line buffer
 *  depending on the type of satellite processed
 */

  int
Detect_Sync_Train( int search_limit )
{
  switch( gbl_sat_type )
  {
	case NOAA   : return( Detect_NOAA_Sync(search_limit) );
	case METEOR : return( Detect_Meteor_Sync(search_limit) );
  }

  return(0);

} /* End of Detect_Sync_Train() */


/*------------------------------------------------------------------------*/

/*  NOAA_Image_Line()
 *
 *  Fills one line of one channel's image by
 *  reading samples from buffer and processing
 */

  int
NOAA_Image_Line( unsigned char *image_line, int *max_gray_val )
{
  unsigned char
	sample_a, /* First of the pair of carrier samples  */
	sample_b; /* Second of the pair of carrier samples */

  int
	pixel_val, /* Un-normalized value of pixel */
	line_idx,  /* Index to image line buffer   */
	idx;       /* Index for loops etc */


  /* Process samples to make one line of image */
  for( line_idx = 0; line_idx < NOAA_IMAGE_WIDTH; line_idx++ )
  {
	pixel_val = 0;

	/*** Summate 3 carrier samples for each pixel value ***/
	for( idx = 0; idx < 3; idx++ )
	{
	  /* Read first of the pair of carrier samples */
	  sample_a = gbl_line_buffer[gbl_buffer_idx];

	  /* If end of samples block reached, return error */
	  if( ++gbl_buffer_idx >= BLOCK_SIZE )
		return(-1);

	  /* Read second of the pair of carrier samples */
	  sample_b = gbl_line_buffer[gbl_buffer_idx];

	  /* This is the instantaneous magnitude of the 2.4 Khz  */
	  /* subcarrier, calculated as the scalar magnitude of   */
	  /* two consecutive audio samples at 90 deg phase angle */
	  pixel_val += (int)Scalar_Mag( sample_a, sample_b );

	} /* End of for( idx = 0; idx < 3; idx++ ) */

	/* Normalize pixel values to below 255 */
	image_line[line_idx] = (pixel_val * NOAA_SCALE_FACTOR) / CARRIER_MAX;

	/* Record max gray value of line */
	if( image_line[line_idx] > *max_gray_val )
	  *max_gray_val = image_line[line_idx];

  } /* End of for( line_idx = 0; line_idx < NOAA_IMAGE_WIDTH; line_idx++ ) */

  return(0);

} /* End of NOAA_Image_Line() */

/*------------------------------------------------------------------------*/

/*  Meteor_Image_Line()
 *
 *  Fills one line of a Meteor APT image by
 *  reading samples from buffer and processing
 */

  int
Meteor_Image_Line( unsigned char *image_line, int *max_gray_val )
{
  unsigned char
	sample_a, /* First of the pair of carrier samples  */
	sample_b; /* Second of the pair of carrier samples */

  int
	pixel_val, /* Un-normalized value of pixel */
	line_idx,  /* Index to image line buffer   */
	idx;       /* Index for loops etc */

  /* Process samples to make one line of image */
  for( line_idx = 0; line_idx < METEOR_IMAGE_WIDTH; line_idx++ )
  {
	pixel_val = 0;

	/*** Summate 5 carrier samples for each pixel value ***/
	for( idx = 0; idx < 5; idx++ )
	{
	  /* Read first of the pair of carrier samples */
	  sample_a = gbl_line_buffer[gbl_buffer_idx];

	  /* If end of samples block reached, return error */
	  if( ++gbl_buffer_idx >= BLOCK_SIZE )
		return(-1);

	  /* Read second of the pair of carrier samples */
	  sample_b = gbl_line_buffer[gbl_buffer_idx];

	  /* This is the instantaneous magnitude of the 2.4 Khz  */
	  /* subcarrier, calculated as the scalar magnitude of   */
	  /* two consecutive audio samples at 90 deg phase angle */
	  pixel_val += (int)Scalar_Mag( sample_a, sample_b );

	} /* End of for( idx = 0; idx < 5; idx++ ) */

	/* Normalize pixel values to below 255 */
	image_line[line_idx] = (pixel_val * METEOR_SCALE_FACTOR) / CARRIER_MAX;

	/* Record max gray value of line */
	if( image_line[line_idx] > *max_gray_val )
	  *max_gray_val = image_line[line_idx];

  } /* End of for( line_idx = 0; line_idx < image_width; line_idx++ ) */

  return(0);

} /* End of Meteor_Image_Line() */

/*------------------------------------------------------------------------*/

/*  Dsp_Process_Image()
 *
 *  Processes APT images direct from DSP audio samples.
 */

  int
Dsp_Process_Image( char image_file[], int duration )
{
  /* Stores error conditions during processing */
  int error_flag = 0;

  /* Process APT images according to satellite type */
  switch( gbl_sat_type )
  {
	case NOAA : /* Decode NOAA APT images */
	  {
		/* Image file names of NOAA ch A and B images */
		char
		  A_image_file[90],
		  B_image_file[90];

		int
		  line_cnt,        /* Number of APT image lines processed       */
		  carrier_val,     /* Instantaneous value of 2.4 Khz subcarrier */
		  A_max_gray_val,  /* Maximum gray value in channel A PGM image */
		  B_max_gray_val,  /* Maximum gray value in channel B PGM image */
		  A_image_height,  /* Height in lines of channel A PGM image    */
		  B_image_height,  /* Height in lines of channel B PGM image    */
		  A_image_size,    /* Current size in bytes of A image buffer   */
		  B_image_size;    /* Current size in bytes of B image buffer   */

		/* Buffers for the output images */
		unsigned char *A_image_buffer;
		unsigned char *B_image_buffer;

		FILE
		  *A_image_fp, /* Channel A image file  */
		  *B_image_fp; /* Channel B image file  */

		/* Process file name */
		File_Name( image_file );

		/* Prepare image file names */
		strcpy( A_image_file, image_file );
		strcpy( B_image_file, image_file );
		strcat( A_image_file, "-NOAA-A.pgm" );
		strcat( B_image_file, "-NOAA-B.pgm" );

		/* Print some information */
		if( duration )
		  fprintf( stderr, "wxapt: Setting decoding duration to %d sec\n",
			  duration );

		fprintf( stderr, "wxapt: Decoding images: %s  %s ...\n",
			A_image_file, B_image_file );

		/* Clear variables before decoding samples buffer */
		A_image_height = B_image_height = 0;
		A_image_size   = B_image_size = 0;
		A_max_gray_val = B_max_gray_val = 0;
		A_image_buffer = B_image_buffer = NULL;
		carrier_val    = line_cnt = error_flag = 0;

		/* Time limit in image lines (2/sec) */
		duration *= 2;

		/* Keep processing samples to make an image */
		while( !gbl_stop )
		{
		  /* Fill APT samples buffer */
		  if( Dsp_Fill_Buffer() == -1 )
		  {
			error_flag = -1;
			break;
		  }

		  /* Bypass sync and space/markers of ch A. */
		  if( (gbl_buffer_idx += NOAA_SYNC_CHA_SPACE) >= BLOCK_SIZE )
		  {
			error_flag = -1;
			break;
		  }

		  /* Re-allocate memory to image buffers, abort on error */
		  A_image_buffer = realloc( A_image_buffer,
			  A_image_size + NOAA_IMAGE_WIDTH);
		  if( A_image_buffer == NULL )
			Abort_On_Error( -17 );

		  B_image_buffer = realloc( B_image_buffer,
			  B_image_size + NOAA_IMAGE_WIDTH);
		  if( B_image_buffer == NULL )
			Abort_On_Error( -17 );

		  /* If line cannot be filled break out */
		  if( NOAA_Image_Line( A_image_buffer + A_image_size,
				&A_max_gray_val) == -1 )
		  {
			error_flag = -1;
			break;
		  }

		  A_image_height++;
		  A_image_size = NOAA_IMAGE_WIDTH * A_image_height;

		  /* Bypass ch A gray scale/ch B sync/ch B space/markers */
		  if( (gbl_buffer_idx += NOAA_CHA_CHB_SPACE) >= BLOCK_SIZE )
		  {
			error_flag = -1;
			break;
		  }

		  /* If line cannot be filled break out */
		  if( NOAA_Image_Line( B_image_buffer + B_image_size,
				&B_max_gray_val) == -1 )
		  {
			error_flag = -1;
			break;
		  }

		  B_image_height++;
		  B_image_size = NOAA_IMAGE_WIDTH * B_image_height;

		  /* Stop if line count reaches time limit (sec*2) */
		  if( ++line_cnt == duration )
			break;

		} /* End of while( !gbl_stop ) */

		/* Carry out histogram normalization if not suppressed */
		if( !(gbl_action & NO_NORMALIZE) )
		{
		  Normalize( A_image_buffer, A_image_size );
		  Normalize( B_image_buffer, B_image_size );
		  A_max_gray_val = B_max_gray_val = 255;
		}

		/* Rotate images 180 degrees if requested */
		if( gbl_action & ROTATE_IMAGE )
		{
		  Rotate( A_image_buffer, A_image_size );
		  Rotate( B_image_buffer, B_image_size );
		}

		/*** Processing finished, complete image files ***/

		/* Open Channel A image file */
		A_image_fp = fopen( A_image_file, "w+" );
		if( A_image_fp == NULL )
		{
		  perror( A_image_file );
		  exit(-1);
		}

		/* Open Channel B image file */
		B_image_fp  = fopen( B_image_file, "w+" );
		if( B_image_fp == NULL )
		{
		  perror( B_image_file );
		  exit(-1);
		}

		/* Write header in Ch-A output PGM files */
		fprintf( A_image_fp, "P5\n" );
		fprintf( A_image_fp, "%d %d\n", NOAA_IMAGE_WIDTH, A_image_height );
		fprintf( A_image_fp, "%d\n",      A_max_gray_val );

		/* Write Ch-A image buffer to file, abort on error */
		if( fwrite(A_image_buffer, 1, A_image_size, A_image_fp) != A_image_size )
		  Abort_On_Error(-15);

		/* Write header in Ch-B output PGM file */
		fprintf( B_image_fp, "P5\n" );
		fprintf( B_image_fp, "%d %d\n", NOAA_IMAGE_WIDTH, B_image_height );
		fprintf( B_image_fp, "%d\n",      B_max_gray_val );

		/* Write Ch-B image buffer to file, abort on error */
		if( fwrite(B_image_buffer, 1, B_image_size, B_image_fp) != B_image_size )
		  Abort_On_Error(-15);

		/* Free buffers and close image files */
		free( A_image_buffer );
		free( B_image_buffer );

		fprintf( stderr, "wxapt: Closing files: %s  %s\n",
			A_image_file, B_image_file );

		fclose( A_image_fp );
		fclose( B_image_fp );

		if( gbl_stop )
		  error_flag = 0;

	  } /* End of case NOAA : */

	  break;

	case METEOR : /* Decode Meteor APT images */
	  {
		/* Image file name of Meteor image */
		char Met_image_file[90];

		int
		  line_cnt,      /* Number of APT image lines processed       */
		  carrier_val,   /* Instantaneous value of 2.4 Khz subcarrier */
		  max_gray_val,  /* Maximum gray value in Meteor APT image    */
		  Met_image_height,  /* Height in lines of Meteor APT image   */
		  Met_image_size;    /* Current size in bytes of Meteor image */

		/* Buffer for the output image */
		unsigned char *Met_image_buffer;

		/* Meteor APT image file pointer */
		FILE *Met_image_fp;

		/* Process file name */
		File_Name( image_file );

		/* Prepare image file name */
		strcpy( Met_image_file, image_file );
		strcat( Met_image_file, "-Meteor.pgm" );

		/* Print some information */
		if( duration )
		  fprintf( stderr, "wxapt: Setting decoding duration to %d sec\n",
			  duration );

		fprintf( stderr, "wxapt: Decoding image: %s ...\n",
			Met_image_file );

		/* Clear variables before decoding sample buffer */
		Met_image_height = max_gray_val = 0;
		Met_image_size   = 0;
		Met_image_buffer = NULL;
		carrier_val = line_cnt = error_flag = 0;

		/* Time limit in image lines (2/sec) */
		duration *= 2;

		/* Keep processing samples to make an image */
		while( !gbl_stop )
		{
		  /* Fill APT samples buffer */
		  if( Dsp_Fill_Buffer() == -1 )
		  {
			error_flag = -1;
			break;
		  }

		  /* Bypass sync and space/markers of image */
		  if( (gbl_buffer_idx += METEOR_SYNC_IMAGE_SPACE) >= BLOCK_SIZE )
		  {
			error_flag = -1;
			break;
		  }

		  /* Re-allocate memory to image buffer, abort on error */
		  Met_image_buffer = realloc( Met_image_buffer,
			  Met_image_size + METEOR_IMAGE_WIDTH);
		  if( Met_image_buffer == NULL )
			Abort_On_Error( -17 );

		  /* Break out if line cannot be filled */
		  if( Meteor_Image_Line( Met_image_buffer + Met_image_size,
				&max_gray_val) == -1 )
		  {
			error_flag = -1;
			break;
		  }

		  Met_image_height++;
		  Met_image_size = METEOR_IMAGE_WIDTH * Met_image_height;

		  /* Stop if line count reaches time limit (sec*2) */
		  if( ++line_cnt == duration )
			break;

		} /* End of while( !gbl_stop ) */

		/* Carry out histogram normalization if not suppressed */
		if( !(gbl_action & NO_NORMALIZE) )
		{
		  Normalize( Met_image_buffer, Met_image_size );
		  max_gray_val = 255;
		}

		/* Rotate images 180 degrees if requested */
		if( gbl_action & ROTATE_IMAGE )
		  Rotate( Met_image_buffer, Met_image_size );

		/*** Processing finished, complete image files ***/

		/* Open Meteor image file */
		Met_image_fp  = fopen( Met_image_file, "w+" );
		if( Met_image_fp == NULL )
		{
		  perror( Met_image_file );
		  exit(-1);
		}

		/* Write header of Meteor output PGM files */
		fprintf( Met_image_fp, "P5\n" );
		fprintf( Met_image_fp, "%d %d\n",
			METEOR_IMAGE_WIDTH, Met_image_height );
		fprintf( Met_image_fp, "%d\n", max_gray_val );

		/* Write image buffer to file, abort on error */
		if( fwrite(Met_image_buffer, 1, Met_image_size, Met_image_fp) !=
			Met_image_size )
		  Abort_On_Error(-15);

		/* Free image buffer and close image file */
		free( Met_image_buffer );
		fprintf( stderr, "wxapt: Closing file: %s\n", Met_image_file );
		fclose( Met_image_fp );

		if( gbl_stop )
		  error_flag = 0;

	  } /* End of case METEOR : */

  } /* End of switch( gbl_sat_type ) */

  return( error_flag );

} /* End of Dsp_Process_Image() */

/*------------------------------------------------------------------------*/

/*  File_Process_Image()
 *
 *  Processes APT images from audio samples file.
 */

  int
File_Process_Image( char sample_file[], char image_file[] )
{
  /* Stores error conditions during processing */
  int error_flag;

  /* File suffix to be removed from file name */
  char suffix[20];

  /* Position of suffix in file name */
  char *suffix_pos;

  /* Process file name */
  File_Name( sample_file );

  fprintf( stderr, "wxapt: Opening APT audio samples file: %s\n",
	  sample_file );

  /* Open APT audio samples file */
  if( (gbl_sample_fp = fopen(sample_file, "r")) == NULL )
  {
	perror( sample_file );
	exit(-1);
  }

  /* Read satellite type from file, abort on error */
  if( (error_flag = fread( &gbl_sat_type, 4, 1, gbl_sample_fp )) == -1 )
  {
	perror( "wxapt: fread()" );
	exit(-1);
  }
  if( error_flag != 1 )
	Abort_On_Error(-16);

  gbl_sync_ref = gbl_sync_refs[ gbl_sat_type ];

  /* Creat the suffix to be removed from sample file name. */
  /* This is to remove the suffix <sat_name>.bin which is  */
  /* added automatically by wxapt to samples file name.    */
  suffix[0] = '-';
  suffix[1] = '\0';
  strcat( suffix, gbl_sat_names[gbl_sat_type] );
  strcat( suffix, ".bin" );

  /* Find position of suffix in file name */
  suffix_pos = strstr( sample_file, suffix );

  /* Cut out suffix if it exists */
  if( suffix_pos != NULL )
	suffix_pos[0] = '\0';

  error_flag = 0;

  /* Process APT images according to satellite type */
  switch( gbl_sat_type )
  {
	case NOAA : /* Decode NOAA APT images */
	  {
		/* Image file names of NOAA ch A and B images */
		char
		  A_image_file[90],
		  B_image_file[90];

		int
		  line_cnt,        /* Number of APT image lines processed       */
		  carrier_val,     /* Instantaneous value of 2.4 Khz subcarrier */
		  A_max_gray_val,  /* Maximum gray value in channel A PGM image */
		  B_max_gray_val,  /* Maximum gray value in channel B PGM image */
		  A_image_height,  /* Height in lines of channel A PGM image    */
		  B_image_height,  /* Height in lines of channel B PGM image    */
		  A_image_size,    /* Current size in bytes of A image buffer   */
		  B_image_size;    /* Current size in bytes of B image buffer   */

		/* Buffers for the output images */
		unsigned char *A_image_buffer;
		unsigned char *B_image_buffer;

		FILE
		  *A_image_fp,   /* Channel A image file  */
		  *B_image_fp;   /* Channel B image file  */

		/* Process file name */
		if( strlen(image_file) == 0 )
		  strcpy( image_file, sample_file );
		else
		  File_Name( sample_file );

		/* Prepare image file names */
		strcpy( A_image_file, image_file );
		strcpy( B_image_file, image_file );
		strcat( A_image_file, "-NOAA-A.pgm" );
		strcat( B_image_file, "-NOAA-B.pgm" );

		/* Print some information */
		fprintf( stderr, "wxapt: Decoding images: %s  %s ...\n",
			A_image_file, B_image_file );

		/* Clear variables before decoding samples buffer */
		A_image_height = B_image_height = 0;
		A_image_size   = B_image_size   = 0;
		A_max_gray_val = B_max_gray_val = 0;
		A_image_buffer = B_image_buffer = NULL;
		carrier_val    = line_cnt = error_flag = 0;

		/* Process samples from file to make an image */
		while( File_Fill_Buffer() != -1 )
		{
		  /* Bypass A sync and space/markers of ch A. */
		  if( (gbl_buffer_idx += NOAA_SYNC_CHA_SPACE) >= BLOCK_SIZE )
		  {
			error_flag = -1;
			break;
		  }

		  /* Re-allocate memory to image buffers, abort on error */
		  A_image_buffer = realloc( A_image_buffer,
			  A_image_size + NOAA_IMAGE_WIDTH);
		  if( A_image_buffer == NULL )
			Abort_On_Error( -17 );

		  B_image_buffer = realloc( B_image_buffer,
			  B_image_size + NOAA_IMAGE_WIDTH);
		  if( B_image_buffer == NULL )
			Abort_On_Error( -17 );

		  if( NOAA_Image_Line( A_image_buffer + A_image_size,
				&A_max_gray_val) == -1 )
		  {
			error_flag = -1;
			break;
		  }

		  A_image_height++;
		  A_image_size = NOAA_IMAGE_WIDTH * A_image_height;

		  /* Bypass ch A gray scale/ch B sync/ch B space/markers */
		  if( (gbl_buffer_idx += NOAA_CHA_CHB_SPACE) >= BLOCK_SIZE )
		  {
			error_flag = -1;
			break;
		  }

		  /* If line cannot be filled break out */
		  if( NOAA_Image_Line( B_image_buffer + B_image_size,
				&B_max_gray_val) == -1 )
		  {
			error_flag = -1;
			break;
		  }

		  B_image_height++;
		  B_image_size = NOAA_IMAGE_WIDTH * B_image_height;

		} /* End of while( File_Fill_Buffer() != -1 ) */

		/* Carry out histogram normalization if not suppressed */
		if( !(gbl_action & NO_NORMALIZE) )
		{
		  Normalize( A_image_buffer, A_image_size );
		  Normalize( B_image_buffer, B_image_size );
		  A_max_gray_val = B_max_gray_val = 255;
		}

		/* Rotate images 180 degrees if requested */
		if( gbl_action & ROTATE_IMAGE )
		{
		  Rotate( A_image_buffer, A_image_size );
		  Rotate( B_image_buffer, B_image_size );
		}

		/*** Processing finished, complete image files ***/

		/* Open Channel A image file */
		A_image_fp  = fopen( A_image_file, "w+" );
		if( A_image_fp == NULL )
		{
		  perror( A_image_file );
		  exit(-1);
		}

		/* Open Channel B image file */
		B_image_fp  = fopen( B_image_file, "w+" );
		if( B_image_fp == NULL )
		{
		  perror( B_image_file );
		  exit(-1);
		}

		/* Write header in Ch-A output PGM files */
		fprintf( A_image_fp, "P5\n" );
		fprintf( A_image_fp, "%d %d\n", NOAA_IMAGE_WIDTH, A_image_height );
		fprintf( A_image_fp, "%d\n", A_max_gray_val );

		/* Write Ch-A image buffer to file, abort on error */
		if( fwrite(A_image_buffer, 1, A_image_size, A_image_fp) != A_image_size )
		  Abort_On_Error(-15);

		/* Write header in Ch-B output PGM file */
		fprintf( B_image_fp, "P5\n" );
		fprintf( B_image_fp, "%d %d\n", NOAA_IMAGE_WIDTH, B_image_height );
		fprintf( B_image_fp, "%d\n", B_max_gray_val );

		/* Write B image buffer to file, abort on error */
		if( fwrite(B_image_buffer, 1, B_image_size, B_image_fp) != B_image_size )
		  Abort_On_Error(-15);

		/* Free image buffers and close image files */
		free( A_image_buffer );
		free( B_image_buffer );

		fprintf( stderr, "wxapt: Closing files: %s  %s\n",
			A_image_file, B_image_file );

		fclose( A_image_fp );
		fclose( B_image_fp );

	  } /* End of case NOAA : */

	  break;

	case METEOR: /* Decode METEOR APT images */
	  {
		/* Meteor image file name */
		char Met_image_file[90];

		int
		  line_cnt,     /* Number of APT image lines processed       */
		  carrier_val,  /* Instantaneous value of 2.4 Khz subcarrier */
		  max_gray_val, /* Maximum gray value of Meteor APT image    */
		  Met_image_height, /* Height in lines of Meteor APT image   */
		  Met_image_size;   /* Current size in bytes of Meteor image */

		/* Buffer for the output image */
		unsigned char *Met_image_buffer;

		/* Meteor image file pointer */
		FILE *Met_image_fp;

		/* Process file name */
		if( strlen(image_file) == 0 )
		  strcpy( image_file, sample_file );
		else
		  File_Name( sample_file );

		/* Prepare image file name */
		strcpy( Met_image_file, image_file );
		strcat( Met_image_file, "-Meteor.pgm" );

		/* Print some information */
		fprintf( stderr, "wxapt: Decoding image: %s\n",
			Met_image_file );

		/* Clear variables before decoding sample buffer */
		Met_image_height = max_gray_val = 0;
		Met_image_size   = 0;
		Met_image_buffer = NULL;
		carrier_val = line_cnt = error_flag = 0;

		/* Process samples from file to make an image */
		while( File_Fill_Buffer() != -1 )
		{
		  /* Bypass sync and space/markers of image */
		  if( (gbl_buffer_idx += METEOR_SYNC_IMAGE_SPACE) >= BLOCK_SIZE )
		  {
			error_flag = -1;
			break;
		  }

		  /* Re-allocate memory to image buffer, abort on error */
		  Met_image_buffer = realloc( Met_image_buffer,
			  Met_image_size + METEOR_IMAGE_WIDTH);
		  if( Met_image_buffer == NULL )
			Abort_On_Error( -17 );

		  /* Break out if line cannot be filled */
		  if( Meteor_Image_Line( Met_image_buffer + Met_image_size,
				&max_gray_val) == -1 )
		  {
			error_flag = -1;
			break;
		  }

		  Met_image_height++;
		  Met_image_size = METEOR_IMAGE_WIDTH * Met_image_height;

		} /* End of while( File_Fill_Buffer() != -1 ) */

		/* Carry out histogram normalization if not suppressed */
		if( !(gbl_action & NO_NORMALIZE) )
		{
		  Normalize( Met_image_buffer, Met_image_size );
		  max_gray_val = 255;
		}

		/* Rotate images 180 degrees if requested */
		if( gbl_action & ROTATE_IMAGE )
		  Rotate( Met_image_buffer, Met_image_size );

		/*** Processing finished, complete image files ***/

		/* Open Meteor APT image file */
		Met_image_fp  = fopen( Met_image_file, "w+" );
		if( Met_image_fp == NULL )
		{
		  perror( Met_image_file );
		  exit(-1);
		}

		/* Write header in output Meteor PGM files */
		fprintf( Met_image_fp, "P5\n" );
		fprintf( Met_image_fp, "%d %d\n",
			METEOR_IMAGE_WIDTH, Met_image_height );
		fprintf( Met_image_fp, "%d\n", max_gray_val );

		/* Write image buffer to file, abort on error */
		if( fwrite(Met_image_buffer, 1, Met_image_size, Met_image_fp) !=
			Met_image_size )
		  Abort_On_Error(-15);

		/* Free image buffer and close image file */
		free( Met_image_buffer );
		fprintf( stderr, "wxapt: Closing file: %s\n",
			Met_image_file );
		fclose( Met_image_fp );

	  } /* End of case METEOR : */

  } /* End of switch( gbl_sat_type ) */

  return( error_flag );

} /* End of File_Process_Image() */

/*------------------------------------------------------------------------*/

/*  Normalize()
 *
 *  Does histogram (linear) normalization of a pgm (P5) image file
 */

  void
Normalize( unsigned char *image_buffer, int image_size )
{
  int
	hist[256],  /* Intensity histogram of pgm image file  */
	blk_cutoff, /* Count of pixels for black cutoff value */
	wht_cutoff, /* Count of pixels for white cutoff value */
	pixel_val,  /* Used for calculating normalized pixels */
	pixel_cnt,  /* Total pixels counter for cut-off point */
	idx;        /* Index for loops etc */

  int
	black_val,  /* Black cut-off pixel intensity value */
	white_val,  /* White cut-off pixel intensity value */
	val_range;  /* Range of intensity values in image  */

  if( image_size <= 0 )
  {
	fprintf( stderr,
		"wxapt: Image file seems empty - normalization not performed\n" );
	return;
  }

  /* Clear histogram */
  for( idx = 0; idx < 256; idx++ )
	hist[ idx ] = 0;

  /* Build image intensity histogram */
  for( idx = 0; idx < image_size; idx++ )
	++hist[ image_buffer[idx] ];

  /* Determine black/white cut-off counts */
  blk_cutoff = (image_size * BLACK_CUT_OFF) / 100;
  wht_cutoff = (image_size * WHITE_CUT_OFF) / 100;

  /* Find black cut-off intensity value */
  pixel_cnt = 0;
  for( black_val = 0; black_val < 256; black_val++ )
  {
	pixel_cnt += hist[ black_val ];
	if( pixel_cnt > blk_cutoff )
	  break;
  }

  /* Find white cut-off intensity value */
  pixel_cnt = 0;
  for( white_val = 255; white_val >= 0; white_val-- )
  {
	pixel_cnt += hist[ white_val ];
	if( pixel_cnt > wht_cutoff )
	  break;
  }

  /* Rescale pixels in image for full intensity range */
  val_range = white_val - black_val;
  if( (int) val_range <= 0 )
  {
	fprintf( stderr,
		"wxapt: Image seems flat - normalization not performed\n" );
	return;
  }

  /* Perform histogram normalization on images */
  fprintf( stderr,
	  "wxapt: Performing histogram normalization on image ...\n" );

  for( pixel_cnt = 0; pixel_cnt < image_size; pixel_cnt++ )
  {
	pixel_val = image_buffer[ pixel_cnt ];
	pixel_val = ( (pixel_val - black_val) * 255 ) / val_range;

	pixel_val = ( pixel_val < 0 ? 0 : pixel_val );
	pixel_val = ( pixel_val > 255 ? 255 : pixel_val );
	image_buffer[ pixel_cnt ] = pixel_val;
  }

} /* End of Normalize() */

/*------------------------------------------------------------------------*/

/*  Rotate()
 *
 *  Rotates a pgm (P5) image by 180 degrees
 */

  void
Rotate( unsigned char *image_buffer, int image_size )
{
  int idx; /* Index for loops etc */

  unsigned char
	*idx_temp,  /* Buffer location to be saved in temp    */
	*idx_swap;  /* Buffer location to be swaped with temp */

  /* Holds a pixel value temporarily */
  unsigned char temp;

  if( image_size <= 0 )
  {
	fprintf(stderr,
		"wxapt: Image file seems empty - rotation not performed");
	return;
  }

  /* Rotate image 180 degrees */
  fprintf( stderr, "wxapt: Rotating image by 180 degrees ...\n" );

  for( idx = 0; idx < image_size / 2; idx++ )
  {
	idx_temp = image_buffer + idx;
	idx_swap = image_buffer - 1 + image_size - idx;
	temp = *( idx_temp );
	*( idx_temp ) = *( idx_swap );
	*( idx_swap ) = temp;
  }

} /* End of Rotate() */

/*------------------------------------------------------------------------*/

/*  Scalar_Mag()
 *
 *  Since sampling of the 2.4 Khz sub-carrier is done at 9.6 KHz
 *  consecutive samples are at 90 deg. The instantaneous magnitude
 *  of the carrier is sqrt(x^2+y^2) were x and y are two
 *  consecutive samples. This is approximated by a Taylor
 *  series expansion in (x+y) to avoid floating point sqrt,
 *  e.g: ((x+y)^2-x*y)/(x+y) to first order, and refined by
 *  one iteration of the Newton-Raphson method. This is rather
 *  wierd but seems to give the most accurate approximation.
 */

  unsigned char
Scalar_Mag( unsigned char x, unsigned char y )
{
  /* Scalar magnitude of two samples at 90 deg */
  unsigned char scalar_mag;

  int
	sum,    /* Sum of two consecutive samples    */
	prod,   /* Product of two cnsecutive samples */
	aprox;  /* First approxim. to sqrt(x^2+y^2)  */


  /* Remove the 'neutral' value (0x80)   */
  /* of the unsigned 8-bit audio samples */
  x = abs(x - 128);
  y = abs(y - 128);

  /* Calculate first approximation to sqrt(x^2+y^2) */
  sum   = x + y;
  prod  = x * y;
  aprox = ( sum == 0 ? 0 : (sum*sum - prod)/sum );

  /* Refine by one iteration using Newton-Raphson method */
  scalar_mag = ( aprox == 0 ? 0 :
	  ((aprox*aprox+sum*sum-(prod*2))/aprox)/2 );

  return( scalar_mag );

} /* End of Scalar_Mag() */

/*------------------------------------------------------------------------*/
