/*  sound.c
 *
 *  Sound-card handling functions of wxapt application
 */

/*
 *  wxapt: An application to decode APT signals from weather
 *  satellites and produce an image of the weather.
 *
 *  Copyright (C) 2001  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */


#include "wxapt.h"

/* Weather satellite type (NOAA/Meteor/etc) */
extern int gbl_sat_type;

/* Satellite type names */
extern char *gbl_sat_names[];

/* Required position of sync train in line buffer */
extern int gbl_sync_ref;

/* List of required sync position references */
extern int gbl_sync_refs[];

extern int
  gbl_dsp_fd,   /* File descriptor of DSP device   */
  gbl_mix_fd;   /* File descriptor of mixer device */

/* Pointer to samples file */
extern FILE *gbl_sample_fp;

/* Audio/DSP samples buffer for one line */
extern unsigned char gbl_line_buffer[];

/* Index to audio samples buffer */
extern int gbl_buffer_idx;

/* Stop flag set by signal handler on user interrupt */
extern int gbl_stop;

/* DSP block (chunk) size */
int gbl_chunk_size;

/* Buffer for reading in DSP data */
unsigned char *gbl_dsp_buffer;
/* Index for this buffer */
int gbl_buff_idx;

/* dsp sampling rate error */
int dsp_rate_err;

/*------------------------------------------------------------------------*/

/*  Setup_Sound_Card()
 *
 *  Sets up mixer and DSP device and adjusts input
 *  level for correct sampling of APT signal
 */

  void
Setup_Sound_Card( void )
{
  int
	rec_src,     /* Recording/Capture source (Line?)  */
	reclev_dev,  /* Recording Level device (IGAIN?)   */
	record_lev,  /* Recording level request to mixer  */
	dsp_speed,   /* DSP sampling speed (9600 Hz)      */
	sample_fmt,  /* DSP sample format, 8-bit unsigned */
	channels,    /* (Stereo) or Mono mode selection   */
	itmp;        /* Temporary int for tests etc       */

  /* Sound device names from souncard.h */
  char *dev_name[SOUND_MIXER_NRDEVICES] = SOUND_DEVICE_NAMES;

  /* File lock information   */
  struct flock lockinfo;


  /*** Open and lock mixer and dsp devices ***/
  fprintf( stderr, "%s\n", "wxapt: Setting up the sound card ..." );

  /* Open mixer device, abort on error */
  if( (gbl_mix_fd = open(MIXER_DEV, O_RDWR, 0)) == -1 )
  {
	perror( "wxapt: "MIXER_DEV );
	exit(-1);
  }

  /* Attempt to lock mixer device */
  fprintf( stderr, "%s\n", "wxapt: Attempting to lock "MIXER_DEV" ..." );

  /* Attempt to lock entire mixer device file */
  lockinfo.l_type   = F_WRLCK;
  lockinfo.l_whence = SEEK_SET;
  lockinfo.l_start  = 0;
  lockinfo.l_len    = 0;

  /* If mixer device is already locked, abort */
  if( fcntl( gbl_mix_fd, F_SETLK, &lockinfo ) < 0 )
  {
	perror( "wxapt: Lock "MIXER_DEV );
	fcntl( gbl_mix_fd, F_GETLK, &lockinfo );
	fprintf( stderr, "wxapt: Lock "MIXER_DEV
		": Device already locked by pid %d%s\n", lockinfo.l_pid,
		" - aborting" );
	exit( -1 );
  }

  /* Open DSP device, abort on error */
  if( (gbl_dsp_fd = open( DSP_DEV, O_RDONLY, 0)) == -1 )
  {
	perror( "wxapt: "DSP_DEV );
	exit(-1);
  }

  /* Attempt to lock DSP device */
  fprintf( stderr, "%s\n", "wxapt: Attempting to lock "DSP_DEV" ..." );

  /* Attempt to lock entire DSP device file */
  lockinfo.l_type   = F_RDLCK;
  lockinfo.l_whence = SEEK_SET;
  lockinfo.l_start  = 0;
  lockinfo.l_len    = 0;

  /* If DSP device is already locked, abort */
  if( fcntl( gbl_dsp_fd, F_SETLK, &lockinfo ) < 0 )
  {
	perror( "wxapt: Lock "DSP_DEV );
	fcntl( gbl_dsp_fd, F_GETLK, &lockinfo );
	fprintf( stderr, "wxapt: Lock "DSP_DEV
		": Device already locked by pid %d%s\n", lockinfo.l_pid,
		" - aborting" );
	exit( -1 );
  }

  /*** Set up sound card parameters ***/

  /* Attempt to set up sound card */
  fprintf( stderr, "%s\n", "wxapt: Setting up sound card ..." );

  /* Find number of recording source matching name in wxapt.h */
  for( rec_src = 0; rec_src < SOUND_MIXER_NRDEVICES; rec_src++ )
	if( strcmp( dev_name[rec_src], RECORD_SRC ) == 0 )
	  break;

  if( rec_src == SOUND_MIXER_NRDEVICES )
	Abort_On_Error(-24);

  /* Make recording source mask */
  itmp = 1 << rec_src;

  /* Attempt to select recording source, abort on error */
  if( ioctl(gbl_mix_fd, SOUND_MIXER_WRITE_RECSRC, &itmp) == -1 )
  {
	perror("wxapt: SOUND_MIXER_WRITE_RECSRC");
	exit(-1);
  }
  if( itmp != 1 << rec_src )
	Abort_On_Error(-3);

  /* Zero line inpu level to silence (my) speakers */
  record_lev = 0;
  if( ioctl(gbl_mix_fd, MIXER_WRITE(rec_src), &record_lev) == -1 )
  {
	perror("wxapt: Setting recording source level");
	exit(-1);
  }
  if( record_lev != 0 )
	Abort_On_Error(-6);

  /* Find number of recording level device matching name in wxapt.h */
  for( reclev_dev = 0; reclev_dev < SOUND_MIXER_NRDEVICES; reclev_dev++ )
	if( strcmp( dev_name[reclev_dev], RECLEV_DEV ) == 0 )
	  break;

  if( reclev_dev ==  SOUND_MIXER_NRDEVICES )
	Abort_On_Error(-25);

  /* Write default level to recording level device, abort on error */
  record_lev = RECORD_LEV;
  if( ioctl(gbl_mix_fd, MIXER_WRITE(reclev_dev), &record_lev) == -1 )
  {
	perror("wxapt: Setting recording level device");
	exit(-1);
  }
  if( record_lev != RECORD_LEV )
	Abort_On_Error(-6);

  /* Setup dsp fragment size to 4k, ignore errors */
  gbl_chunk_size = (NUM_OF_CHANNELS << 16) + 12;
  ioctl( gbl_dsp_fd, SNDCTL_DSP_SETFRAGMENT, &gbl_chunk_size );

  /* Set sample size to unsigned 8-bit, abort on error */
  sample_fmt = AFMT_U8;
  if( ioctl(gbl_dsp_fd, SNDCTL_DSP_SETFMT, &sample_fmt) == -1 )
  {
	perror( "wxapt: SNDCTL_DSP_SETFMT");
	exit(-1);
  }
  if( sample_fmt != AFMT_U8 )
	Abort_On_Error(-4);

  /* Set stereo or mono mode */
  channels = NUM_OF_CHANNELS;
  if( ioctl(gbl_dsp_fd, SNDCTL_DSP_CHANNELS, &channels) == -1 )
	perror("SNDCTL_DSP_CHANNELS");

  /* Setup DSP sampling speed, abort on error */
  dsp_speed = DSP_SPEED;
  if( ioctl(gbl_dsp_fd, SNDCTL_DSP_SPEED, &dsp_speed) == -1 )
  {
	perror( "wxapt: SNDCTL_DSP_SPEED");
	exit(-1);
  }
  if( dsp_speed != DSP_SPEED )
	Abort_On_Error(-5);

  /* Read DSP block (chunck) size */
  if( ioctl(gbl_dsp_fd, SNDCTL_DSP_GETBLKSIZE, &gbl_chunk_size ) == -1 )
  {
	perror("wxapt: Reading DSP block (chunk) size");
	exit(-1);
  }

  /* Allocate memory to DSP data buffer */
  gbl_dsp_buffer = malloc( gbl_chunk_size );
  if( gbl_dsp_buffer == NULL )
	Abort_On_Error( -17 );

  fprintf( stderr, "%s\n", "wxapt: Sound card setup completed" );

} /* End of Setup_Sound_Card() */

/*------------------------------------------------------------------------*/

/*  Setup_Audio_Level()
 *
 *  Displays a simple graphical aid for setting
 *  up the audio level from the APT receiver
 */

  void
Setup_Audio_Level()
{
  /* Used for simple graphical  */
  /* indication of record level */
  char graph[57];

  int
	carrier_max, /* Max value of carrier in  buffer */
	carrier_val, /* Sampled value of 2.4kHz carrier */
	level_error, /* Error in the input audio level  */
	idx;         /* Index used in loops etc         */


  fprintf( stderr, "\nDisplaying error in audio input level setting" );
  fprintf( stderr, "\nExit via User Interrupt (Ctrl-C) when finished\n" );

  /*** Loop until Input audio level is correct ***/
  while(!gbl_stop)
  {
	/* Fill in the samples buffer from DSP */
	if( Read_Dsp_Buffer(gbl_line_buffer, BLOCK_SIZE) == -1 )
	  Abort_On_Error( -1 );

	/*** Find max value of 2.4kHz subcarrier in buffer ***/
	carrier_max = 0;

	for( idx = 0; idx < BLOCK_SIZE-1; idx++ )
	{
	  carrier_val = Scalar_Mag(gbl_line_buffer[idx],
		  gbl_line_buffer[idx+1]);

	  /* Find maximum carrier level */
	  if( carrier_val > carrier_max )
		carrier_max = carrier_val;
	}

	/* Find input audio level error and keep in limits */
	level_error = carrier_max - CARRIER_MAX;
	if( level_error > +24 )
	  level_error = +24;
	if( level_error < -24 )
	  level_error = -24;

	/* Display graph of max audio level */
	strcpy( graph,
		"LO<----+----+----+---->----0----<----+----+----+---->HI");

	graph[level_error+27] = 'x';
	fprintf( stderr, "%s%c", graph, 0x0d );

  } /* End of while(!gbl_stop) */

} /* Setup_Audio_Level() */

/*------------------------------------------------------------------------*/

/*  Synchronize_Buffer()
 *
 *  Synchronizes samples buffer with the APT sync train
 */

  int
Synchronize_Buffer(void)
{
  int
	sync_err,     /* Sync position error relative to refer   */
	ave_sync_err, /* Average of sync train position error    */
	block_size;   /* Data block size used as read() argument */

  char graph[8*SYNC_DETECT+2];
  int idx;

  /* Print some information */
  fprintf( stderr, "%s\n",
	  "wxapt: Synchronizing samples buffer with APT sync train" );

  /* Try to detect satellite type from sync train signature */
  fprintf( stderr, "wxapt: Looking for a %s type sync train ...\n",
	  gbl_sat_names[gbl_sat_type] );

  gbl_sync_ref = gbl_sync_refs[ gbl_sat_type ];

  /* Fill in the line buffer from DSP */
  gbl_buffer_idx = 0;
  if( Read_Dsp_Buffer(gbl_line_buffer, BLOCK_SIZE) == -1 )
	return( -1 );

  /* Set buffer index to just after sync train */
  Detect_Sync_Train( BLOCK_SIZE );
  sync_err = gbl_buffer_idx - gbl_sync_ref;

  /* To avoid false sync detection */
  ave_sync_err = 1;

  do /* Try to synchronize with APT sync train */
  {
	if( gbl_stop )
	  exit( 0 ); /* User interrupt */

	/* Coarse adjustment of the sync train position in the */
	/* samples buffer is done by dummy-reading dsp samples */
	block_size = sync_err;
	if( block_size < 0 )
	  block_size += BLOCK_SIZE;
	block_size += (block_size * dsp_rate_err) / BLOCK_SIZE;

	/* Fill in the samples buffer from DSP. This is a 'dummy' read */
	/* to dump most of the samples before the sync train position  */
	Read_Dsp_Buffer( gbl_line_buffer, block_size );

	/* Fill in the samples buffer from DSP */
	Read_Dsp_Buffer( gbl_line_buffer, BLOCK_SIZE + dsp_rate_err );

	/* Set buffer index to just after sync train */
	gbl_buffer_idx = 0;
	Detect_Sync_Train( BLOCK_SIZE );

	/* Error in sync train position relative to ref */
	sync_err = gbl_buffer_idx - gbl_sync_ref;

	/* If close to sync, start dsp rate error measurement */
	if( abs(sync_err) < SYNC_DETECT )
	{
	  if( sync_err > 0 )
		dsp_rate_err++;
	  else
		if( sync_err < 0 )
		  dsp_rate_err--;

	  /* Window average of sync error */
	  ave_sync_err = ( 3*ave_sync_err + 4*sync_err ) / 4;

	  /* Plot a graph of sync error */
	  idx = ave_sync_err + 4*SYNC_DETECT;
	  strcpy( graph, "|+++|+++|+++|+++|+++|+++|+++|+++0+++|+++|+++|+++|+++|+++|+++|+++|" );
	  graph[ idx ] = 'x';
	  fprintf( stderr, "%s%c", graph, 0x0d );
	}

  } /* End of do */

  /* Continue if sync error not zero */
  while( abs(ave_sync_err) != 0 );

  /* Print some information */
  fprintf( stderr, "\n%s\n",
	  "wxapt: Samples buffer synchronized with APT sync train");

  return(0);

} /* End of Synchronize_Buffer() */

/*------------------------------------------------------------------------*/

/*  Record_APT_Signal()
 *
 *  Records samples of the 2.4 Khz subcarrier in a file
 */

  void
Record_APT_Signal( char sample_file[], int duration )
{
  /* Number of APT image lines processed */
  int line_cnt;

  /* Count of samples returned by read()/write() */
  ssize_t sample_cnt;

  /* Prepare sound card */
  Setup_Sound_Card();

  /* Synchronize buffer, abort on error */
  if( Synchronize_Buffer() == -1 )
	Abort_On_Error(-12);

  /* Process file name */
  File_Name( sample_file );

  /* Add extension to file name */
  strcat( sample_file, "-" );
  strcat( sample_file, gbl_sat_names[gbl_sat_type] );
  strcat( sample_file, ".bin" );

  /* Create a file for writng samples */
  gbl_sample_fp = fopen( sample_file, "w+" );
  if( gbl_sample_fp == NULL )
  {
	perror( sample_file );
	exit(-1);
  }

  /* Read samples from DSP, stop on user interrupt */
  /* or at time limit as number of APT lines read. */
  /* duration is in seconds of recording time.   */
  if( duration )
	fprintf( stderr,
		"wxapt: Setting recording duration to %d sec\n",
		duration );

  /* Print some information */
  fprintf( stderr,
	  "wxapt: Recording APT audio signal samples in %s\n",
	  sample_file);

  line_cnt = 0;
  duration *= 2; /* Time limit in image lines (2/sec) */

  /* Write satellite type to samples file, abort on error */
  if( (sample_cnt = fwrite( &gbl_sat_type, 4, 1, gbl_sample_fp )) == -1 )
  {
	perror( "wxapt: fwrite()" );
	exit(-1);
  }
  if( sample_cnt != 1 )
	Abort_On_Error(-2);

  while( !gbl_stop )
  {
	/* Fill samples buffer from DSP */
	Dsp_Fill_Buffer();

	/* Write carrier samples to samples file, abort on error */
	sample_cnt = fwrite(gbl_line_buffer, 1, BLOCK_SIZE, gbl_sample_fp );
	if( (sample_cnt == -1) && (!gbl_stop) )
	{
	  perror( "wxapt: fwrite()" );
	  exit(-1);
	}
	if( (sample_cnt != BLOCK_SIZE) )
	  Abort_On_Error(-8);

	/* Stop if line count reaches time limit (sec*2) */
	if( ++line_cnt == duration )
	  break;

  } /* End of while( !gbl_stop ) */

  /* Print some information */
  fprintf( stderr, "wxapt: Closing samples file: %s\n", sample_file );

  fclose( gbl_sample_fp );
  if( gbl_dsp_fd > 0) close( gbl_dsp_fd );
  if( gbl_mix_fd > 0) close( gbl_mix_fd );

  return;

} /* End of Record_APT_Signal() */

/*------------------------------------------------------------------------*/

/*  File_Fill_Buffer()
 *
 *  Fills the audio sample buffer from source file
 */

  int
File_Fill_Buffer( void )
{
  /* Count of samples returned by read() */
  int sample_cnt;

  gbl_buffer_idx = 0;

  /* Fill in the samples buffer from file, return -1 on error */
  if( (sample_cnt = fread( gbl_line_buffer, 1,
		  BLOCK_SIZE, gbl_sample_fp )) == -1 )
  {
	perror( "wxapt: fread()" );
	exit(-1);
  }
  if( sample_cnt != BLOCK_SIZE )
	return(-1);

  /* Set buffer index just after sync train.  */
  /* Search for sync up to +30 from sync ref. */
  Detect_Sync_Train( gbl_sync_refs[gbl_sat_type] + 30 );

  /* If sync detection is far out, keep */
  /* buffer index in default position   */
  if( abs(gbl_buffer_idx - gbl_sync_ref) > SYNC_DETECT )
	gbl_buffer_idx = gbl_sync_ref;

  return(0);

} /* End of File_Fill_Buffer() */

/*------------------------------------------------------------------------*/

/*  Dsp_Fill_Buffer()
 *
 *  Fills the audio sample buffer from sound card DSP.
 *  Buffer filling is synchronized with APT sync train.
 */

  int
Dsp_Fill_Buffer( void )
{
  int
	sync_err,   /* Difference between sync ref and pos */
	sync_flag;  /* Sync train detected flag */

  /* Data block size used as argument to read() */
  static int block_size = BLOCK_SIZE;


  /* Fill in the samples buffer from DSP */
  gbl_buffer_idx = 0;
  if( Read_Dsp_Buffer(gbl_line_buffer, block_size+dsp_rate_err ) == -1 )
	return( -1 );

  /* Set buffer index to just after sync train.     */
  /* Search for sync up to +30 from sync reference. */
  gbl_buffer_idx = 0;
  Detect_Sync_Train( gbl_sync_refs[gbl_sat_type] + 30 );
  sync_flag = ( gbl_buffer_idx != 0 );

  /* Try to keep APT sync train at beginning of buffer. This  */
  /* done done by increasing or decreasing the data  block    */
  /* size argument passed to read(), by adding or subtracting */
  /* half the difference between detected sync position and a */
  /* reference value. Error is limited to +-1 to avoid coarse */
  /* corrections to sync position due to noise interference.  */
  sync_err = gbl_buffer_idx - gbl_sync_ref;
  if( sync_err > 2 )
	sync_err = 2;
  else
	if( sync_err < -2 )
	  sync_err = -2;

  gbl_buffer_idx = gbl_sync_ref + sync_err;

  /* Vary the data block size argument to read() to   */
  /* take up any drift in sync position due to errors */
  /* in the DSP sampling rate and satellite movement. */
  if( sync_flag ) /* Sync train detected */
  {
	block_size = BLOCK_SIZE + gbl_buffer_idx - gbl_sync_ref;
	if( block_size > BUFFER_SIZE )
	  block_size = BUFFER_SIZE;
  }

  return(0);

} /* End of Dsp_Fill_Buffer() */

/*------------------------------------------------------------------------*/

/* Read_Dsp_Buffer()
 *
 * Reads the DSP's buffer and separates Left/Right
 * channel data depending on stereo or mono mode
 */

  int
Read_Dsp_Buffer( unsigned char *buffer, int buff_size )
{
  int
	sample_cnt,
	max = 0,
	idx;
  static int flag=1;

  if( flag )
  {
	flag = 0;

	/* Fill in the samples buffer from DSP, return -1 on error */
	if( (sample_cnt = read(gbl_dsp_fd, gbl_dsp_buffer, gbl_chunk_size)) == -1 )
	  if( !gbl_stop )
	  {
		perror( "wxapt: read()" );
		return( -1 );
	  }

	if( sample_cnt != gbl_chunk_size )
	  return( -1 );

	/* Start buffer index according to stereo|mono mode */
	if( NUM_OF_CHANNELS == 1 ) /* Mono */
	  gbl_buff_idx = 0;
	else
	  gbl_buff_idx = CHANNEL_IN_USE;
  }

  /* Transfer data from dsp buffer to buffer to be processed, */
  /* separating left or right channel data if in stereo mode  */
  for( idx = 0; idx < buff_size; idx++ )
  {
	if( gbl_buff_idx >= gbl_chunk_size )
	{
	  /* Fill in the samples buffer from DSP, return -1 on error */
	  if( (sample_cnt = read(gbl_dsp_fd, gbl_dsp_buffer, gbl_chunk_size)) == -1 )
		if( !gbl_stop )
		{
		  perror( "wxapt: read()" );
		  return( -1 );
		}

	  if( sample_cnt != gbl_chunk_size )
		return( -1 );

	  /* Start buffer index according to stereo|mono mode */
	  if( NUM_OF_CHANNELS == 1 ) /* Mono */
		gbl_buff_idx = 0;
	  else
		gbl_buff_idx = CHANNEL_IN_USE;

	} /* if( buff_index >= gbl_chunk_size ) */

	/* Fill buffer from left or right channel. gbl_buff_idx */
	/* is arranged to point to the correct location     */
	buffer[ idx ] = gbl_dsp_buffer[ gbl_buff_idx ];

	if( max < buffer[ idx ] - 128 )
	  max = buffer[ idx ] - 128;

	/* Increment buffer index according to stereo|mono mode */
	gbl_buff_idx += NUM_OF_CHANNELS;

  } /* for( idx = 0; idx < buff_size; idx++ ) */

  return( 0 );

} /* End of Read_Dsp_Buffer() */

/*------------------------------------------------------------------------*/
