/*  wxapt.c
 *
 *  Main body of wxapt application
 */

/*
 *  wxapt: An application to decode APT signals from weather
 *  satellites and produce an image of the weather.
 *
 *  Copyright (C) 2001  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include "wxapt.h"

/* Weather satellite type (NOAA/Meteor/etc) */
int gbl_sat_type;

/* Satellite type names */
char *gbl_sat_names[] = { SAT_NAMES };

/* Required position of sync train in line buffer */
int gbl_sync_ref;

/* List of required sync position references */
int gbl_sync_refs[] = { SYNC_REFS };

int
  gbl_action,   /* Action to carry out according to options */
  gbl_dsp_fd,   /* File descriptor of DSP device   */
  gbl_mix_fd;   /* File descriptor of mixer device */

/* Pointer to samples file */
FILE *gbl_sample_fp;

/* Index to audio sample buffer */
int gbl_buffer_idx;

/* Signals end of operation on user interrupt */
int gbl_stop = 0;

/* Signal handler */
static void sig_handler( int signal );

/*------------------------------------------------------------------------*/

  int
main( int argc, char *argv[] )
{
  int
	start_hrs,  /* Hours part of time to start operation   */
	start_min,  /* Minutes part of time to start operation */
	duration,   /* Time limit of recording or image processing  */
	sleep_sec=0,/* Sleep time in sec till start of operation    */
	option;     /* Command line option returned by getopt()     */

  /* Sample file name */
  char sample_file[71];

  /* Image file name */
  char image_file[71];

  /* getopt() variables */
  extern char *optarg;
  extern int optind, opterr, optopt;

  /* New and old actions for sigaction() */
  struct sigaction sa_new, sa_old;

  /* Initialize new actions */
  sa_new.sa_handler = sig_handler;
  sigemptyset( &sa_new.sa_mask );
  sa_new.sa_flags = 0;

  /* Register function to handle signals */
  sigaction( SIGINT,  &sa_new, &sa_old );
  sigaction( SIGSEGV, &sa_new, 0 );
  sigaction( SIGFPE,  &sa_new, 0 );
  sigaction( SIGTERM, &sa_new, 0 );
  sigaction( SIGABRT, &sa_new, 0 );


  /*** Process command line options ***/

  image_file[0] = sample_file[0] = '\0';
  duration      = 0;
  gbl_sat_type  = NOAA;
  gbl_action    = DEFAULT_ACTION;

  /* If no arguments, follow default action */
  if( argc == 1 )
  {
	Default_Action( image_file, duration );
	return(0);
  }

  /* Process command line options */
  while( (option = getopt(argc, argv, "a:i:p:r::R::s:t:hlnov") ) != -1 )
	switch( option )
	{
	  case 'a' : /* Specify satellite APT type */
		gbl_sat_type  = atoi( optarg );
		break;

	  case 'i' : /* Specify image file name */
		if( strlen(optarg) > 70 )
		  Abort_On_Error(-14);
		strcpy( image_file, optarg );
		break;

	  case 'p' : /* Process a recorded APT signal */
		if( strlen(optarg) > 70 )
		  Abort_On_Error(-14);
		strcpy( sample_file, optarg );
		gbl_action |= PROCESS_FILE;
		break;

	  case 'r' : /* Only record APT signal */
		if( optarg != 0 )
		{
		  if( strlen(optarg) > 70 )
			Abort_On_Error(-14);
		  strcpy( sample_file, optarg );
		}
		gbl_action |= RECORD_ONLY;
		break;

	  case 'R' : /* Record and then process APT signal */
		if( optarg != 0 )
		{
		  if( strlen(optarg) > 70 )
			Abort_On_Error(-14);
		  strcpy( sample_file, optarg );
		}
		gbl_action |= RECORD_PROCESS;
		break;

	  case 's' : /* Set start and end time of operation */
		Start_Stop_Time( &start_hrs, &start_min, &sleep_sec, &duration );
		break;

	  case 't' : /* Set time limit of record/process */
		duration = atoi( optarg );
		if( duration <= 0 )
		  Abort_On_Error(-11);
		/* If time limit is less than 60 sec, assume its in min */
		if( duration < 60 )
		  duration *= 60;
		break;

	  case 'l' : /* Go into audio level setup loop */
		Setup_Sound_Card();
		Setup_Audio_Level();
		return(0);

	  case 'n' : /* Suppress histogram normalization */
		gbl_action |= NO_NORMALIZE;
		break;

	  case 'o' : /* Rotate image over 180 deg */
		gbl_action |= ROTATE_IMAGE;
		break;

	  case 'h' : /* Print usage and exit */
		Usage();
		return(0);

	  case 'v' : /* Print version */
		fprintf( stderr, "%s\n", VERSION );
		return(0);

	  default: /* Print usage and exit */
		Usage();
		exit(-1);

	} /* End of switch( option ) */

  /* Sleep if delay > 0, exit on interrupt */
  if( sleep_sec > 0 )
  {
	fprintf( stderr, "\nwxapt: Going to sleep till %02d:%02d\n",
		start_hrs, start_min );
	sleep( sleep_sec );
  }
  if( gbl_stop )
	exit( SIGINT );

  /* Carry out action according to options */
  switch( gbl_action & 0x3F )
  {
	case DEFAULT_ACTION :
	  Default_Action( image_file, duration );
	  return(0);

	case RECORD_ONLY :
	  Record_APT_Signal( sample_file, duration );
	  return(0);

	case RECORD_PROCESS :
	  Record_APT_Signal( sample_file, duration );
	  if( File_Process_Image( sample_file, image_file ) == -1)
		Abort_On_Error(-13);
	  return(0);

	case PROCESS_FILE :
	  if( File_Process_Image( sample_file, image_file ) == -1)
		Abort_On_Error(-13);
	  return(0);

  } /* End of switch( gbl_action ) */

  return(0);

} /* End of main() */

/*------------------------------------------------------------------------*/

/*  Default_Action()
 *
 *  Carries out wxapt's default action if no arguments.
 *  This is real-time processing of APT signal into images.
 */

  void
Default_Action( char image_file[], int duration )
{
  /* Prepare sound card */
  Setup_Sound_Card();

  /* Synchronize buffer, abort on error */
  if( Synchronize_Buffer() == -1 )
	Abort_On_Error(-12);

  /* Process APT samples direct from */
  /* DSP and produce image files.    */
  if( Dsp_Process_Image(image_file, duration) == -1)
	Abort_On_Error(-13);

  /* Wrap up and quit */
  if( gbl_dsp_fd > 0 ) close( gbl_dsp_fd );
  if( gbl_mix_fd > 0 ) close( gbl_mix_fd );

  return;

} /* End of Default_Action() */

/*------------------------------------------------------------------------*/

/*  File_Name()
 *
 *  Prepare a file name from optarg, use date and time if null
 */

  void
File_Name( char *file_name )
{
  int
	len, /* String length of file_name */
	idx; /* Index to while loop */

  idx = 0;

  /* If file_name is null, use date and time as file name */
  if( strlen(file_name) == 0 )
  {
	/* Variables for reading time (UTC) */
	time_t tp;
	struct tm utc;

	/* Prepare a file name as UTC date-time */
	time( &tp );
	utc = *gmtime( &tp );
	strftime( file_name, 17, "%d%b%Y-%H%M", &utc );
  }
  else /* Remove leading spaces from file_name */
  {
	len = strlen(file_name);
	do idx++;
	while( (file_name[idx] == ' ') && (idx != len) );
	file_name += idx - 1;
  }

} /* End of File_Name() */

/*------------------------------------------------------------------------*/

/*  Start_Stop_Time()
 *
 *  Calculates sleep time and duration of recording or decoding.
 *  If optarg for the -s option is only 5 char, this is taken as
 *  the time to start recording or decoding and the number of
 *  sec for wxapt to sleep is calculated. If optarg is 11 char
 *  long then sleep time and duration of recording or decoding
 *  is also calculated. The expected format is -s<hh:mm[-hh:mm]>
 */

void
Start_Stop_Time(
	int *start_hrs, int *start_min,
	int *sleep_sec, int *duration )
{
  extern char *optarg;

  int
	stop_hrs,   /* Stop time hours   */
	stop_min,   /* Stop time minutes */
	stop_sec,   /* Stop time in sec since 00:00 hrs  */
	start_sec,  /* Start time in sec since 00:00 hrs */
	time_sec,   /* Time now in sec since 00:00 hrs   */
	optarg_len; /* Length of optarg  */

  /* Used to read real time */
  struct tm time_now;
  time_t t;

  /* Test optarg length (5 or 11 char) */
  optarg_len = strlen( optarg );
  if( !((optarg_len == 5) || (optarg_len == 11)) )
	Abort_On_Error(-19);

  /* Extract start time and sleep duration. */
  /* (Expected format hh:mm in 24-hour UTC) */
  if( optarg[2] !=':' )
	Abort_On_Error(-19);

  /* Extract hours and minutes of start time */
  *start_hrs = atoi( optarg );
  if( (*start_hrs < 0) || (*start_hrs > 23) )
	Abort_On_Error(-19);

  *start_min = atoi( &optarg[3] );
  if( (*start_min < 0) || (*start_min > 59) )
	Abort_On_Error(-19);

  /* Time now */
  t = time( &t );
  time_now = *gmtime( &t );
  time_sec = time_now.tm_hour*3600 +
	time_now.tm_min*60 +
	time_now.tm_sec;

  start_sec  = *start_hrs*3600 + *start_min*60;
  *sleep_sec = start_sec - time_sec;

  /* If stop time is also specified, extract */
  /* stop time and duration of record/decode */
  if( optarg_len == 11 )
  {
	/* Check basic format ( Expected is (hh:mm)-hh:mm ) */
	if( (optarg[5] != '-') || (optarg[8] != ':') )
	  Abort_On_Error(-19);

	stop_hrs = atoi( &optarg[6] );
	if( (stop_hrs < 0) || (stop_hrs > 23) )
	  Abort_On_Error(-19);

	stop_min = atoi( &optarg[9] );
	if( (stop_min < 0) || (stop_min > 59) )
	  Abort_On_Error(-19);

	stop_sec = stop_hrs*3600 + stop_min*60;
	stop_sec -= time_sec;

	if( (*sleep_sec < 0) && (stop_sec > 0) )
	{
	  fprintf( stderr,
		  "wxapt: Start time has elapsed - starting now\n" );
	  *sleep_sec = 0;
	}

	if( stop_sec < 0 )
	  stop_sec += 86400; /* Next day */

	*duration = stop_sec - *sleep_sec;

	if( *duration <= 0 )
	  Abort_On_Error( -20 );

  } /* if( optarg_len == 11 ) */

  if( *sleep_sec < 0 )
	*sleep_sec += 86400; /* Next day */

} /* End of Start_Stop_Time() */

/*------------------------------------------------------------------------*/

/*  Usage()
 *
 *  Prints usage information
 */

  void
Usage(void)
{
  fprintf( stderr, "%s\n",
	  "Usage: wxapt [-a<satellite-type>] [-s<start-time[-<stop-time>]>] [-t<duration>]");

  fprintf( stderr, "%s\n\n",
	  "             [-i<file>] [-p<file>] [-r[<file>]] [-R[<file>]] [-lnohv]");

  fprintf( stderr, "%s\n\n",
	  "       -a<satellite-type>: 0=NOAA   1=METEOR");

  fprintf( stderr, "%s\n\n",
	  "       -s<start-time[-<stop-time>]>: Time (UTC) in format hh:mm at which\n"
	  "       to start and optionally to stop APT recording or processing");

  fprintf( stderr, "%s\n\n",
	  "       -t<duration>: Duration of APT recording or processing. If duration\n"
	  "       is >= 60, it is taken as seconds, else as minutes");

  fprintf( stderr, "%s\n\n",
	  "       -i<file>: Specify image file name. If this option is not used,\n"
	  "       use date and time as file name: ddMonyy-hhmm<extension>");

  fprintf( stderr, "%s\n\n",
	  "       -p<file>: Process images from samples recorded in <file>");

  fprintf( stderr, "%s\n\n",
	  "       -r[<file>]: Only record APT signal in <file>. If no file name is\n"
	  "       specified, use date and time as file name: ddMonyy-hhmm.bin");

  fprintf( stderr, "%s\n\n",
	  "       -R[<file>]: Record APT signal in <file> and process images\n"
	  "       If no file is specified, files will be named as above");

  fprintf( stderr, "%s\n\n",
	  "       -l: Start in audio input-level setup mode");

  fprintf( stderr, "%s\n\n",
	  "       -n: Suppress histogram normalization of images");

  fprintf( stderr, "%s\n\n",
	  "       -o: Rotate images by 180 degrees");

  fprintf( stderr, "%s\n\n",
	  "       -h: Print this usage information and exit");

  fprintf( stderr, "%s\n\n",
	  "       -v: Print version number and exit");

} /* End of Usage() */

/*------------------------------------------------------------------------*/

/*  sig_handler()
 *
 *  Signal Action Handler function
 */

static void sig_handler( int signal )
{
  fprintf( stderr, "\n" );
  switch( signal )
  {
	case SIGINT :
	  gbl_stop = 1;
	  fprintf( stderr, "%s\n", "wxapt: Exiting via User Interrupt" );
	  break;

	case SIGSEGV :
	  fprintf( stderr, "%s\n", "wxapt: Segmentation Fault" );
	  exit( signal );

	case SIGFPE :
	  fprintf( stderr, "%s\n", "wxapt: Floating Point Exception" );
	  exit( signal );

	case SIGABRT :
	  fprintf( stderr, "%s\n", "wxapt: Abort Signal received" );
	  exit( signal );

	case SIGTERM :
	  fprintf( stderr, "%s\n", "wxapt: Termination Request received" );
	  exit( signal );
  }

} /* End of sig_handler() */

/*------------------------------------------------------------------------*/

/*  Abort_On_Error()
 *
 *  Prints an error message and exits
 */

  void
Abort_On_Error( int why )
{
  /* Don't print error messages if wxapt is stopped */
  if( !gbl_stop )
	switch( why )
	{
	  fprintf( stderr, "\n" );
	  case -1 : /* Abort if samples buffer cannot be filled */
		fprintf( stderr, "%s\n",
			"wxapt: Error filling audio samples buffer - aborting" );
		break;

	  case -2 : /* Abort satellite type cannot be written to file */
		fprintf( stderr, "%s\n",
			"wxapt: Writing satellite type to file failed - aborting" );
		break;

	  case -3 : /* Abort if recording source selection fails */
		fprintf( stderr, "%s\n",
			"wxapt: Recording source selection failed - aborting" );
		break;

	  case -4 : /* Abort if sample size setting fails */
		fprintf( stderr, "%s\n",
			"wxapt: Sample size setting failed - aborting" );
		break;

	  case -5 : /* Abort if recording speed setting fails */
		fprintf( stderr, "%s\n",
			"wxapt: Recording speed setting failed - aborting" );
		break;

	  case -6 : /* Abort if recording level setting fails */
		fprintf( stderr, "%s\n",
			"wxapt: Recording level setting failed - aborting" );
		break;

	  case -7 : /* Abort if samples buffer cannot be filled */
		fprintf( stderr, "%s\n",
			"wxapt: Recording buffer not filled by read() - aborting" );
		break;

	  case -8 : /* Abort if samples buffer cannot be written to file */
		fprintf( stderr, "%s\n",
			"wxapt: Error writing recording buffer to file - aborting" );
		break;

	  case -9 : /* Abort if audio input level too high */
		fprintf( stderr, "%s\n",
			"wxapt: Audio input too high to set recording level - aborting" );
		break;

	  case -10 : /* Abort if audio input level too low */
		fprintf( stderr, "%s\n",
			"wxapt: Audio input too low to set recording level - aborting" );
		break;

	  case -11 : /* Abort if record/process time limit looks bad */
		fprintf( stderr, "%s\n",
			"wxapt: Record/Process duration bad - aborting" );
		Usage();
		break;

	  case -12 : /* Abort if synchronization of buffer fails */
		fprintf( stderr, "%s\n",
			"wxapt: Buffer synchronization with APT sync failed - aborting" );
		break;

	  case -13 : /* Abort if end of samples buffer reached prematurely */
		fprintf( stderr, "%s\n",
			"wxapt: End of samples buffer reached prematurely - aborting" );
		break;

	  case -14 : /* Abort if file name too long */
		fprintf( stderr, "%s\n",
			"wxapt: File name too long - aborting" );
		break;

	  case -15 : /* Abort if image buffer could not be written to file */
		fprintf( stderr, "%s\n",
			"wxapt: Image buffer could not be written to file - aborting" );
		break;

	  case -16 : /* Abort if satellite type cannot be read from file */
		fprintf( stderr, "%s\n",
			"wxapt: Satellite type cannot be read from file - aborting" );
		break;

	  case -17 : /* Abort if malloc() for image buffer fails */
		fprintf( stderr, "%s\n",
			"wxapt: Call to malloc() for a buffer failed - aborting" );
		break;

	  case -18 : /* Abort if normalized image could not be written to file */
		fprintf( stderr, "%s\n",
			"wxapt: Normalized image could not be written to file - aborting" );
		break;

	  case -19 : /* Abort if time to start/stop operation is malformed/invalid */
		fprintf( stderr, "%s\n",
			"wxapt: Time to start or stop is malformed/invalid - aborting" );
		Usage();
		break;

	  case -20 : /* Abort if time to stop operation is invalid */
		fprintf( stderr, "%s\n",
			"wxapt: Stop time is less than or equal to start time - aborting" );
		Usage();
		break;

	  case -21 : /* Abort if no recording level device found */
		fprintf( stderr, "%s\n",
			"wxapt: No recording level device found - aborting" );
		break;

	  case -22 : /* Abort if Line level setting fails */
		fprintf( stderr, "%s\n",
			"wxapt: Line level setting failed - aborting" );
		break;

	  case -23 : /* Abort if Stereo/Mono mode setting fails */
		fprintf( stderr, "%s\n",
			"wxapt: Stereo/Mono mode setting failed - aborting");
		break;

	  case -24 : /* Abort if audio input source name is bad */
		fprintf( stderr, "%s\n",
			"wxapt: Recording source name (RECORD_SRC) is bad - aborting");
		break;

	  case -25 : /* Abort if recording level device name is bad */
		fprintf( stderr, "%s\n",
			"wxapt: Recording level device name (RECLEV_DEV) is bad - aborting");

	}  /* switch( why ) */

  exit( why );

} /* End of Abort_On_Error() */

/*------------------------------------------------------------------------*/

