/*  wxapt.h
 *
 *  Header file for wxapt: An application to decode APT signals
 *  from weather satellites and produce an image of the weather.
 */

/*
 *  wxapt: An application to decode APT signals from weather
 *  satellites and produce an image of the weather.
 *
 *  Copyright (C) 2001  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */


#ifndef WXAPT_H
#define WXAPT_H 1

#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <stdio.h>
#include <time.h>

/************ CUSTOMIZE THESE DEFINITIONS FOR YOUR NEEDS *************/

#define MIXER_DEV      "/dev/mixer" /* Your mixer device */
#define DSP_DEV          "/dev/dsp" /* Your DSP device   */
#define NUM_OF_CHANNELS           2 /* Stereo/Mono: 1=Mono, 2=Stereo  */
#define CHANNEL_IN_USE         LEFT /* Channel for record: LEFT|RIGHT */
#define RECORD_SRC           "line" /* Sound recording source */
#define RECLEV_DEV          "igain" /* Recording level device */
#define RECORD_LEV           0x6060 /* Default level of recording source  */

/***************** END OF CUSTOMIZABLE SECTION ***********************/

/* Generel definitions for image processing */
#define DSP_SPEED         9600 /* DSP recording speed in samples/sec */
#define CARRIER_MAX        126 /* Max 2.4kHz carrier-level required  */
#define BUFFER_SIZE       4900 /* Audio buffer size=samples/line+100 */
#define BLOCK_SIZE        4800 /* Read() block size=samples/line     */
#define SYNC_DETECT          8 /* Sync-detected threshold (samples)  */
#define MAX_FILE_NAME       70 /* Max length for optional filenames  */
#define BLACK_CUT_OFF        1 /* Black cut-off percentile for normalization */
#define WHITE_CUT_OFF        2 /* White cut-off percentile for normalization */

/* Definitions for NOAA image processing */
#define NOAA_IMAGE_WIDTH          696 /* Width in pixels of images produced */
#define NOAA_SYNC_CHA_SPACE       128 /* Space between ch A sync and image  */
#define NOAA_CHA_CHB_SPACE        310 /* Space between ch A & B, in samples */
#define NOAA_FIFO_SIZE             63 /* Size of sync detector samples fifo */
#define NOAA_SYNC_LOWER_THRESH   1800 /* Sync train detector lower threshold*/
#define NOAA_SYNC_UPPER_THRESH   2500 /* Sync train detector upper threshold*/
#define NOAA_SYNC_REF             150 /* Required position of sync train    */
#define NOAA_SCALE_FACTOR          90 /* Factor to keep pixel values < 255  */

/* Definitions for Meteor image processing */
#define METEOR_IMAGE_WIDTH        768 /* Width in pixels of images produced */
#define METEOR_SYNC_IMAGE_SPACE   130 /* Space between sync and image start */
#define METEOR_FIFO_SIZE           40 /* Size of sync detector samples fifo */
#define METEOR_SYNC_LOWER_THRESH 1500 /* Sync train detector lower threshold*/
#define METEOR_SYNC_UPPER_THRESH 2200 /* Sync train detector upper threshold*/
#define METEOR_SYNC_REF           130 /* Required position of sync train    */
#define METEOR_SCALE_FACTOR        46 /* Factor to keep pixel values < 255  */

/* Actions according to command line options */
#define DEFAULT_ACTION           0x00 /* Process images directly from DSP   */
#define RECORD_ONLY              0x01 /* Record snd o/p of receiver to file */
#define RECORD_PROCESS           0x02 /* Record snd to file, process images */
#define PROCESS_FILE             0x04 /* Process images from a recorded file*/
#define ROTATE_IMAGE             0x40 /* Rotate processed images by 180 deg */
#define NO_NORMALIZE             0x80 /* Do not normalize images by default */

/* Satellite type definitions */
#define NOAA                      0 /* NOAA type APT format   */
#define METEOR                    1 /* METEOR type APT format */
#define NUM_SATS                  2 /* Number of APT formats supported */
#define SAT_NAMES   "NOAA", "Meteor"
#define SYNC_REFS   NOAA_SYNC_REF, METEOR_SYNC_REF

/* Channel to be used in audio capture */
#define LEFT                      0 /* Left channel  */
#define RIGHT                     1 /* Right channel */

/* wxapt version string */
#define VERSION  "wxapt 1.3"

/*-------------------------------------------------------------------*/

/* Function prototypes produced by cproto */
/* image.c */
int Detect_NOAA_Sync(int search_limit);
int Detect_Meteor_Sync(int search_limit);
int Detect_Sync_Train(int search_limit);
int NOAA_Image_Line(unsigned char *image_line, int *max_gray_val);
int Meteor_Image_Line(unsigned char *image_line, int *max_gray_val);
int Dsp_Process_Image(char image_file[], int duration);
int File_Process_Image(char sample_file[], char image_file[]);
void Normalize(unsigned char *image_buffer, int image_size);
void Rotate(unsigned char *image_buffer, int image_size);
unsigned char Scalar_Mag(unsigned char x, unsigned char y);
/* sound.c */
void Setup_Sound_Card(void);
void Setup_Audio_Level(void);
int Synchronize_Buffer(void);
void Record_APT_Signal(char sample_file[], int duration);
int File_Fill_Buffer(void);
int Dsp_Fill_Buffer(void);
int Read_Dsp_Buffer(unsigned char *buffer, int buff_size);
/* wxapt.c */
int main(int argc, char *argv[]);
void Default_Action(char image_file[], int duration);
void File_Name(char *file_name);
void Start_Stop_Time(int *start_hrs, int *start_min, int *sleep_sec, int *duration);
void Usage(void);
void Abort_On_Error(int why);

/*-------------------------------------------------------------------*/

#endif /* End of wxapt.h */

